/*
 *  ulock.c
 *
 *  Copyright (C) 2001:  Hubertus Franke (frankeh@watson.ibm.com) 
 *                       IBM Coorporation
 */

#include <time.h>
#include <linux/ulocks.h>

static unsigned long cycles_per_usec = 0;

void ulock_spin_calibration(void)
{
	/* figure out how many loops we can execute per micro second */
	ulock_t ulock = { 0 };
	int count;
	unsigned long loops;
	unsigned long loops_per_round = 10000;
	unsigned long clock1, clock2, clockterm;
	double d_calib;

	count = 0;
	clock1 = clock() ;
	clockterm = clock1 + CLOCKS_PER_SEC/10;
	do {
		loops = loops_per_round;
		while (((volatile)ulock.status <= 0) && loops) loops--;
		clock2 = clock();
		count++;
	} while (clock2 < clockterm);

	loops = loops_per_round * count;

	d_calib = ((double)loops*CLOCKS_PER_SEC) / 
	                   ((double)(clock2-clock1) * 1000000.0);
	cycles_per_usec = (unsigned long) d_calib;
}
	


static __inline__
unsigned long __ulocks_usecs_to_cycles(unsigned long usecs)
{
	return usecs * cycles_per_usec;
}

int __ulock_down_spin(ulock_t *ulock, unsigned long usecs)
{
	/* in previous fastpath:
	 *    status <= 0 or cmpxchg update failed 
	 */
	unsigned long loops = __ulocks_usecs_to_cycles(usecs);
	
	do {
		long status;
		
		/* monitor status without modification to avoid thrashing */
		while (((status = ((volatile)(ulock->status))) <= 0) && loops)
			loops--;
		/* try again */
		if ((status > 0) && 
		    (__compare_and_swap(&ulock->status,status,status-1)==status))
		{
			inc_stat(ulock,ULOCK_STAT_WRITE_SPIN_SUCCESS);
			return 0;
		}
		if (!loops) {
			if (!__ulock_down(ulock)) {
				inc_stat(ulock,ULOCK_STAT_WRITE_SPIN_SUCCESS);
				return 0;
			}
			return ulock_wait(ulock,0);
		}
	} while (1);
}

#if 0
/* this doesn't work, it has race conditions for 3 or more tasks.
 * Trust me it took me a while to figure this out.
 * The <kernel signal clear> code below can be outpaced by 2 other
 * tasks one which got the lock and another one going into kernel
 * just to find the token on the kernel semaphore and continue, thus
 * yielding two tasks holding the lock.
 * We hence have to opt for a solution that avoids the race condition.
 *
 * I leave this here for now so no one proposes this as a better solution :-)
 */

int __ulock_down_spin(ulock_t *ulock, unsigned long usecs)
{

	unsigned long loops = __ulocks_usecs_to_cycles(usecs);
	do {
		/* correct previous try */
		if (!__ulock_up(ulock)) 
			ulock_wait(ulock,0); /* a kernel signal to much */

		/* monitor status without modification to avoid thrashing */
		while (((volatile)ulock->status <= 0) && loops) 
			loops--;
		/* try again */
		if (!__ulock_down(ulock)) {
			inc_stat(ulock,ULOCK_STAT_WRITE_SPIN_SUCCESS);
			return 0;
		}
		if (!loops) {
			inc_stat_safe(ulock,ULOCK_STAT_WRITE_FAILED);
			return ulock_wait(ulock,0);
		}
	} while (1);
}

int usema_down_spin(ulock_t *ulock, long usecs)   /* from ulocks.h */
{
	if (!__ulock_down(ulock)) {
		inc_stat(ulock,ULOCK_STAT_WRITE_SUCCESS);
		return 0;
	}
	return __ulock_down_spin(ulock,usecs);
}

#endif
