/*
 *  usema.c
 *
 *  Copyright (C) 2001:  Hubertus Franke (frankeh@us.ibm.com) 
 *                       IBM Coorporation
 */

/* implements the SysV version of the Exclusive Locks */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>
#include <linux/ulocks.h>

int P_lock(ulock_t *ulock)
{
	int rc;
	int val;
	struct sembuf mysembuf1;

#ifdef LOCK_STATISTICS
	if (!__ulock_down(ulock))
		inc_stat(ulock,ULOCK_STAT_WRITE_SUCCESS);
	else 
		inc_stat_safe(ulock,ULOCK_STAT_WRITE_FAILED);
#endif
	mysembuf1.sem_num = 0;
	mysembuf1.sem_op  = -1;
	mysembuf1.sem_flg = SEM_UNDO;
	rc = semop (ulock->type, &mysembuf1, 1);
	return rc;
}

int P_lock_spin(ulock_t *ulock, unsigned long usecs)
{
	return -ENOSYS;
}

int V_lock(ulock_t *ulock)
{
	int rc;
	struct sembuf mysembuf1;

	mysembuf1.sem_num = 0;
	mysembuf1.sem_op  = 1;
	mysembuf1.sem_flg = SEM_UNDO;
	
#ifdef LOCK_STATISTICS
	__ulock_up(ulock);
#endif
	rc = semop (ulock->type, &mysembuf1, 1);
	return rc;
}

/* if you have time provide a multiple-reader single writer implementation
 * over SYSV locks.
 */

int readlock(ulock_t *ulock)    { return ENOSYS; }
int readunlock(ulock_t *ulock)  { return ENOSYS; }
int writelock(ulock_t *ulock)   { return ENOSYS; }
int writeunlock(ulock_t *ulock) { return ENOSYS; }


static ulock_t *larray;
static int      numulocks;
	
int init_locks(ulock_t* ulocks[], int numlocks, int shared, 
	       int memtype, unsigned int nummobjs, int *lockimpl)
{
	int i;

	*lockimpl = 0; // kernel locks 
	if (shared) { 
		errno = ENOSYS; 
		return -1; 
	}

	numulocks = numlocks;
	larray = (ulock_t*)get_shared_memory(numlocks*sizeof(ulock_t),'s',0);

	for ( i=0 ; i<numlocks ; i++) {
		int sem = semget (IPC_PRIVATE,1,IPC_CREAT | IPC_EXCL | 0660);
		if (sem != -1) {
			ulocks[i] = &larray[i];
			ulocks[i]->type   = sem;
			ulocks[i]->status = 0;
			INIT_COUNTERS(ulocks[i]);
			V_lock(ulocks[i]); /* make lock available */
			continue;
		}
		return -1;
	}
	return 0;
}


int release_locks(void)
{
	int i,rc;
	for ( i=0 ; i<numulocks; i++) {
		rc = semctl(larray[i].type, 0, IPC_RMID, 0);
	}
	return rc;
}





