/*
 *  my_ulock.c
 *
 *  Copyright (C) 2001:  Hubertus Franke (frankeh@watson.ibm.com) 
 *                       IBM Coorporation
 */

#include <fcntl.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <asm/cache.h>
#include <errno.h>
#include "setup.h"
#include <linux/ulocks.h>

static _numlocks = 0;

void *get_shared_memory(int size, int type, int idx);
void release_shared_memory();

int init_locks(ulock_t* ulocks[], int numlocks, int shared, 
	       int memtype, unsigned int nummobjs, int *lockimpl)
{
	int i,rc1,rc2;
	void* addrs[MAX_MEMOBJS];
	ulock_t *larray;
	int segsize;

	*lockimpl = 1; // user locks
#if 0
	/* first test availability of the kernel function */
	{
		ulock_t testlock = INITIALIZED_USEMA(1);
		if ((ulock_create(&testlock)) == -1)
			return -1;
		ulock_destroy(&testlock);
	}
#endif

	ulock_spin_calibration();

	if (nummobjs > MAX_MEMOBJS) 
		nummobjs = MAX_MEMOBJS;
	segsize = (numlocks+nummobjs-1) / nummobjs;
	for (i=0;i<nummobjs;i++) {
		addrs[i] = get_shared_memory(segsize*sizeof(ulock_t),
					     memtype,i);
		if (addrs[i] == NULL) {
			printf("E_%d> %s\n",errno,strerror(errno));
			return -1;
		}
	}
	for ( i=0 ; i < numlocks ; i++) {
		larray   = (ulock_t*)(addrs[i%nummobjs]);
		ulocks[i]  = (ulock_t*)&larray[i/nummobjs];
		if (shared) 
			urwlock_rp_init(ulocks[i]);
		else
			usema_init(ulocks[i],1);
	}
	_numlocks = numlocks;
	return 0;
}

int release_locks(void)
{
}

int P_lock(ulock_t *ulock)
{
	return usema_down(ulock);
}

int P_lock_spin(ulock_t *ulock, unsigned long usecs)
{
	return usema_down_spin(ulock,usecs);
}

int V_lock(ulock_t *ulock)
{
	return usema_up(ulock);
}

	
int readlock(ulock_t *ulock)
{
	return urwlock_readlock(ulock);
}

int readunlock(ulock_t *ulock, int id)
{
	return urwlock_readunlock(ulock);
}

int writelock(ulock_t *ulock, int id)
{
	return urwlock_writelock(ulock);
}

int writeunlock(ulock_t *ulock, int id)
{
	return urwlock_writeunlock(ulock);
}

